/*
 * mcp2515.h
 *
 * Created: 16.08.2012 20:30:59
 *  Author: ls4446s
 */ 


#ifndef MCP2515_H_
#define MCP2515_H_

//--------------------------------------------------------------------------------------------------------
// DATA STRUCTURE
typedef struct  
{
	uint16_t id;
	uint8_t rtr;
	uint8_t length;
	uint8_t data[8];
} CANMessage;

//--------------------------------------------------------------------------------------------------------
// buffer registers
// transmit buffer control registers
#define TXB0CTRL		0x30
#define TXB1CTRL		0x40
#define TXB2CTRL		0x50
#define ABTF			6
#define MLOA			5
#define TXERR			4
#define TXREQ			3
#define TXP1			1
#define TXP0			

// TXnRTS pin control and status register
#define TXRTSCTRL		0x0D
#define B2RTS			5
#define B1RTS			4
#define B0RTS			3
#define B2RTSM			2
#define B1RTSM			1
#define B0RTSM			0

// TXBnSIDH transmit buffer n standard identifier high
#define TXB0SIDH		0x31
#define TXB1SIDH		0x41
#define TXB2SIDH		0x51

// TXBnSIDL	transmit buffer n standard identifier low
#define TXB0SIDL		0x32
#define TXB1SIDL		0x42
#define TXB2SIDL		0x52
#define EXIDE			3

// TXBnEID8 transmit buffer n extended identifier high
#define TXB0EID8		0x33
#define TXB1EID8		0x43
#define TXB2EID8		0x53

// TXBnEID0 transmit buffer n extended identifier low
#define TXB0EID0		0x34
#define TXB1EID0		0x44
#define TXB2EID0		0x54

// TXBnDLC transmit buffer n data length code
#define TXB0DLC			0x35
#define TXB1DLC			0x45
#define TXB2DLC			0x55
#define RTR				6
#define DLC3			3
#define DLC2			2
#define DLC1			1
#define DLC0			0

// TXBnDm transmit buffer n data byte m
#define TXB0D0			0x36
#define TXB0D1			0x37
#define TXB0D2			0x38
#define TXB0D3			0x39
#define TXB0D4			0x3A
#define TXB0D5			0x3B
#define TXB0D6			0x3C
#define TXB0D7			0x3D

#define TXB1D0			0x46
#define TXB1D1			0x47
#define TXB1D2			0x48
#define TXB1D3			0x49
#define TXB1D4			0x4A
#define TXB1D5			0x4B
#define TXB1D6			0x4C
#define TXB1D7			0x4D

#define TXB2D0			0x56
#define TXB2D1			0x57
#define TXB2D2			0x58
#define TXB2D3			0x59
#define TXB2D4			0x5A
#define TXB2D5			0x5B
#define TXB2D6			0x5C
#define TXB2D7			0x5D

// RXB0CTRL receive buffer 0 control register
#define RXB0CTRL		0x60
#define RXM1			6
#define RXM0			5
#define RXRTR			3
#define BUKT			2
#define BUKT1			1
#define FILHIT0			0

// RXB1CTRL receive buffer 1 control register
#define RXB1CTRL		0x70
#define FILHIT2			2
#define FILHIT1			1

// BFPCTRL - RXnBF pin control status register
#define BFPCTRL			0x0C
#define B1BFS			5
#define B0BFS			4
#define B1BFE			3
#define B0BFE			2
#define B1BFM			1
#define B0BFM			0

// RCBnSIDH	receive buffer n standard identifier high registers
#define RXB0SIDH		0x61
#define RXB1SIDH		0x71

// RXBnSIDL receive buffer n standard identifier low registers
#define RXB0SIDL		0x62
#define RXB1SIDL		0x72
#define IDE				3

// RXBnEID8 receive buffer n extended identifier high registers
#define RXB0EID8		0x63
#define RXB1EID8		0x73

// RXBnEID0 receive buffer n extenden identifier low registers
#define RXB0EID0		0x64
#define RXB1EID0		0x74

// RXBnDLC receive buffer n data length code registers
#define RXB0DLC			0x65
#define RXB1DLC			0x75
#define RB1				5
#define RB2				4

// RXBnDm receive buffer n data byte m
#define RXB0D0			0x66
#define RXB0D1			0x67
#define RXB0D2			0x68
#define RXB0D3			0x69
#define RXB0D4			0x6A
#define RXB0D5			0x6B
#define RXB0D6			0x6C
#define RXB0D7			0x6D

#define RXB1D0			0x76
#define RXB1D1			0x77
#define RXB1D2			0x78
#define RXB1D3			0x79
#define RXB1D4			0x7A
#define RXB1D5			0x7B
#define RXB1D6			0x7C
#define RXB1D7			0x7D

#define RXM0SIDH		0x20
#define RXM0SIDL		0x21
#define RXM0EID8		0x22
#define RXM0EID0		0x23
#define RXM1SIDH		0x24
#define RXM1SIDL		0x25
#define RXM1EID8		0x26
#define RXM1EID0		0x27
//--------------------------------------------------------------------------------------------------------
// INTERRUPTS
// CANINTE interrupt enable register
#define CANINTE			0x2B
#define MERRE			7
#define WAKIE			6
#define ERRIE			5
#define TX2IE			4
#define TX1IE			3
#define TX0IE			2
#define RX1IE			1
#define RX0IE			0

// CANINTF interrupt flag register
#define CANINTF			0x2C
#define MERRF			7
#define WAKIF			6
#define ERRIF			5
#define TX2IF			4
#define TX1IF			3
#define TX0IF			2
#define RX1IF			1
#define RX0IF			0

//---------------------------------------------------------------------------------------------------------
// BIT TIMING

// CNF1 configuration 1
#define CNF1			0x2A
#define BRP5			5
#define BRP4			4
#define BRP3			3
#define BRP2			2
#define BRP1			1
#define BRP0			0

// CNF2 configuration 2
#define CNF2			0x29
#define BTLMODE			7
#define PHSEG11			4

// CNF3 configuration 3
#define CNF3			0x28
#define PHSEG21			1

// CAN Bitrate 125 kbps
#define R_CNF1  (1<<BRP2)|(1<<BRP1)|(1<<BRP0)
#define R_CNF2  (1<<BTLMODE)|(1<<PHSEG11)
#define R_CNF3  (1<<PHSEG21)

//---------------------------------------------------------------------------------------------------------
// ERROR DETECTION
// TEC transmit error counter
#define TEC				0x1C

// REC receive error counter
#define REC				0x1D

// EFLG	error flag
#define EFLG			0x2D
#define RX1OVR			7
#define RX0OVR			6
#define TXBO			5
#define TXEP			4
#define RXEP			3
#define TXWAR			2
#define RXWAR			1
#define EWARN			0

//---------------------------------------------------------------------------------------------------------
// MODES OF OPERATION
// CANCTRL can control register
#define CANCTRL			0x0F
#define REQOP2			7
#define REQOP1			6
#define REQOP0			5
#define ABAT			4
#define OSM				3
#define CLKEN			2
#define CLKPRE1			1
#define CLKPRE0			0

// CANSTAT can status register
#define CANSTAT			0x0E
#define OPMOD2			7
#define OPMOD1			6
#define OPMOD0			5
#define ICOD2			3
#define ICOD1			2
#define ICOD0			1

//---------------------------------------------------------------------------------------------------------
// SPI INSTRUCTION SET
#define CAN_Reset			0b11000000
#define CAN_Read			0b11
#define CAN_ReadRxBuffer	0b10010000
#define CAN_Write			0b10
#define CAN_LoadTx			0b1000000
#define CAN_RTS				0b10000000
#define CAN_ReadStatus		0b10100000
#define CAN_RxStatus		0b10110000
#define CAN_BitModify		0b101

//---------------------------------------------------------------------------------------------------------
// MCP2515 FUNCTIONS
void MCP2515_SoftwareReset()
{
	SS_Low();
	SPI_MasterTransmitAndReceive(CAN_Reset);
	_delay_ms(1);
	SS_High();
}

void MCP2515_ReadRegister(uint8_t *storage,uint8_t start_adress, uint8_t number_of_bytes_to_read)
{
	SS_Low();
	SPI_MasterTransmitAndReceive(CAN_Read);
	SPI_MasterTransmitAndReceive(start_adress);
	for (uint8_t i=0;i<number_of_bytes_to_read;i++)
		*(storage+i)=SPI_MasterTransmitAndReceive(0xff);
	SS_High();
}

/*
// reads Rx 0,1
// values for readbuffer:
// 0b000	for buffer 0 starting at RXB0SIDH
// 0b010	for buffer 0 starting at RXB0D0
// 0b100	for buffer 1 starting at RXB1SIDH
// 0b110	for buffer 1 starting at RXB1D0*/
void MCP2515_ReadRx(uint8_t readbuffer,CANMessage *p_message)
{
	SS_Low();
	SPI_MasterTransmitAndReceive(CAN_ReadRxBuffer|readbuffer);
	if (readbuffer&(1<<1))
		for (uint8_t i=0;i<10;i++)	// macht keinen SINN, da CANMessage nur 8 BYTE speichert.
			p_message->data[i]=SPI_MasterTransmitAndReceive(0xff);
	else
		for (uint8_t k=0;k<8;k++)
			p_message->data[k]=SPI_MasterTransmitAndReceive(0xff);
	SS_High();
}

/*
// writes Tx 0,1,2
// values for writebuffer:
// 0b000	TXB0 starting at TXB0SIDH
// 0b001	TXB0 starting at TXB0D0
// 0b010	TXB1 starting at TXB1SIDH
// 0b011	TXB1 starting at TXB1D0
// 0b100	TXB2 starting at TXB2SIDH
// 0b101	TXB2 starting at TXB2D0*/
void MCP2515_LoadTx(uint8_t writebuffer,CANMessage *p_message)
{
	SS_Low();
	SPI_MasterTransmitAndReceive(CAN_LoadTx|writebuffer);
	for (uint8_t i=0;i<p_message->length;i++)
		SPI_MasterTransmitAndReceive(p_message->data[i]);
	SS_High();
}

void MCP2515_WriteRegister(uint8_t adress, uint8_t data)
{
	SS_Low();
	SPI_MasterTransmitAndReceive(CAN_Write);
	SPI_MasterTransmitAndReceive(adress);
	SPI_MasterTransmitAndReceive(data);
	SS_High();
}

/*
// values for sendbuffer:
// 0b000	MCP2515 ignores command
// 0b001	sends TXB0
// 0b010	sends TXB1
// 0b100	sends TXB2
// 0b011    sends TXB0, TXB1
// 0b101	sends TXB0, TXB2
// 0b110	sends TXB1, TXB2
// 0b111	sends TXB0, TXB1, TXB2*/
void CAN_RequestToSend(uint8_t sendbuffer)
{
	SS_Low();
	SPI_MasterTransmitAndReceive(CAN_RTS|sendbuffer);
	SS_High();
}

void MCP2515_ReadStatus(uint8_t *status)
{
	SS_Low();
	SPI_MasterTransmitAndReceive(CAN_ReadStatus);
	*status=SPI_MasterTransmitAndReceive(0xff);
	SS_High();
}

void MCP2515_ReadRxStatus(uint8_t *status)
{
	SS_Low();
	SPI_MasterTransmitAndReceive(CAN_RxStatus);
	*status=SPI_MasterTransmitAndReceive(0xff);
	SS_High();
}

void MCP2515_BitModify(uint8_t adress,uint8_t mask, uint8_t data)
{
	SS_Low();
	SPI_MasterTransmitAndReceive(CAN_BitModify);
	SPI_MasterTransmitAndReceive(adress);
	SPI_MasterTransmitAndReceive(mask);
	SPI_MasterTransmitAndReceive(data);
	SS_High();
}

void MCP2515_Init()
{
	SPI_MasterInit();
	
	// Software Reset des MCP (gleichbedeutend mit dem ResetPin)
	PORT_SPI&=~(1<<SPI_SS);
	SPI_MasterTransmitAndReceive(CAN_Reset);
	_delay_ms(1);
	PORT_SPI|=(1<<SPI_SS);
	_delay_ms(10);
	
	/*MCP2515_SoftwareReset();
	_delay_ms(10);*/

	/*
     *  Einstellen des Bit Timings
     * 
     *  Fosc       = 16MHz
     *  BRP        = 7                (teilen durch 8)
     *  TQ = 2 * (BRP + 1) / Fosc  (=> 1 uS)
     * 
     *  Sync Seg   = 1TQ
     *  Prop Seg   = (PRSEG + 1) * TQ  = 1 TQ
     *  Phase Seg1 = (PHSEG1 + 1) * TQ = 3 TQ
     *  Phase Seg2 = (PHSEG2 + 1) * TQ = 3 TQ
     * 
     *  Bus speed  = 1 / (Total # of TQ) * TQ
     *             = 1 / 8 * TQ = 125 kHz
     */
	MCP2515_WriteRegister(CNF1,(1<<BRP0)|(1<<BRP1)|(1<<BRP2));
	MCP2515_WriteRegister(CNF2,(1<<BTLMODE)|(1<<PHSEG11));
	MCP2515_WriteRegister(CNF3,(1<<PHSEG21));
	
	// interrupts aktivieren
	MCP2515_WriteRegister(CANINTE,(1<<RX0IE)|(1<<RX1IE));
	
	// RX0,RX1 alle Nachrichten empfangen
	MCP2515_WriteRegister(RXB0CTRL,(1<<RXM0)|(1<<RXM1));
	MCP2515_WriteRegister(RXB1CTRL,(1<<RXM1)|(1<<RXM0));
	
	// alle bits der Empfangsmaske lschen um alles zu empfangen
	MCP2515_WriteRegister(RXM0SIDH,0);
	MCP2515_WriteRegister(RXM0SIDL,0);
	MCP2515_WriteRegister(RXM0EID8,0);
	MCP2515_WriteRegister(RXM0EID0,0);
	
	MCP2515_WriteRegister(RXM1SIDH,0);
	MCP2515_WriteRegister(RXM1SIDL,0);
	MCP2515_WriteRegister(RXM1EID8,0);
	MCP2515_WriteRegister(RXM1EID0,0);
	
	// deaktivieren der Pins RXnBF Pins (High Impedance State)
	MCP2515_WriteRegister(BFPCTRL,0);
	// TXnRTS Bits als Inputs schalten
	MCP2515_WriteRegister(TXRTSCTRL,0);
	// normal operation mode
	MCP2515_BitModify(CANCTRL,0xE0,0);
}


/*
// return-Werte:
	// 0 -> erfolgreicher Sendevorgang via TXB0
	// 1 -> erfolgreicher Sendevorgang via TXB1
	// 2 -> erfolgreicher Sendevorgang via TXB2
	// 0xff -> nicht erfolgreicher Sendevorgang*/
uint8_t CAN_SendMessage(CANMessage *p_message)
{
	uint8_t length = p_message->length;
	uint8_t reg=0;
	
	// berprfung, ob TXB0 verfgbar ist:
	MCP2515_ReadRegister(&reg,TXB0CTRL,1);
	if (reg&(1<<TXREQ))
	{
		// ID einstellen
		MCP2515_WriteRegister(TXB0SIDH,(uint8_t)(p_message->id>>3));
		MCP2515_WriteRegister(TXB0SIDL,(uint8_t)(p_message->id<<5));
		// wenn Nachricht ein Remote Transmit Request ist
		if (p_message->rtr)
		{
			// eine rtr hat zwar eine Lnge, aber keine Daten
			// Nachrichten Lnge +  RTR einstellen
			MCP2515_WriteRegister(TXB0DLC,(1<<RTR)|length);
		}
		else
		{
			// Nachrichten Lnge einstellen
			MCP2515_WriteRegister(TXB0DLC,length);
			// Daten
			for (uint8_t i=0;i<length;i++)
				MCP2515_WriteRegister(TXB0D0+i,p_message->data[i]);
		}
		// CAN Nachricht verschicken
		SS_Low();
		SPI_MasterTransmitAndReceive(CAN_RTS|0x01);
		SS_High();
		
		return 0;
	}
	
	reg=0;
	// berprfung, ob TXB1 verfgbar ist:
	MCP2515_ReadRegister(&reg,TXB1CTRL,1);
	if (reg&(1<<TXREQ))
	{
		MCP2515_WriteRegister(TXB1SIDH,(uint8_t)(p_message->id>>3));
		MCP2515_WriteRegister(TXB1SIDL,(uint8_t)(p_message->id<<5));
		if (p_message->rtr)
			MCP2515_WriteRegister(TXB1DLC,(1<<RTR)|length);
		else
		{
			MCP2515_WriteRegister(TXB1DLC,length);
			for (uint8_t i=0;i<length;i++)
				MCP2515_WriteRegister(TXB1D0+i,p_message->data[i]);
		}
		SS_Low();
		SPI_MasterTransmitAndReceive(CAN_RTS|0x02);
		SS_High();
		
		return 1;
	}
	
	reg=0;
	// berprfung, ob TXB2 verfgbar ist:
	MCP2515_ReadRegister(&reg,TXB2CTRL,1);
	if (reg&(1<<TXREQ))
	{
		MCP2515_WriteRegister(TXB2SIDH,(uint8_t)(p_message->id>>3));
		MCP2515_WriteRegister(TXB2SIDL,(uint8_t)(p_message->id<<5));
		if (p_message->rtr)
			MCP2515_WriteRegister(TXB2DLC,(1<<RTR)|length); 
		else
		{
			MCP2515_WriteRegister(TXB2DLC,length);
			for (uint8_t i=0;i<length;i++)
				MCP2515_WriteRegister(TXB2D0+i,p_message->data[i]);
		}
		SS_Low();
		SPI_MasterTransmitAndReceive(CAN_RTS|0x03);
		SS_High();
		
		return 3;
	}
	return 0xff;
}

uint8_t MCPP2515_ReadRxStatus(void)
{
	uint8_t data;
	SS_Low();
	SPI_MasterTransmitAndReceive(CAN_RxStatus);
	data=SPI_MasterTransmitAndReceive(0xff);
	SS_High();
	return data;	
}

uint8_t CAN_ReceiveMessage(CANMessage *p_message)
{
	uint8_t status;
	MCP2515_ReadRxStatus(&status);
	// Nachricht in Puffer 0
	if (status&(1<<6))
	{
		SS_Low();
		SPI_MasterTransmitAndReceive(CAN_ReadRxBuffer);
	}
	// Nachricht in Puffer 1
	else if (status&(1<<7))
	{
		SS_Low();
		SPI_MasterTransmitAndReceive(CAN_ReadRxBuffer|0x04);
	}
	else
		return 0xff;	// FEHLER
	p_message->id=(uint16_t)SPI_MasterTransmitAndReceive(0xff)<<3;
	p_message->id|=(uint16_t)SPI_MasterTransmitAndReceive(0xff)>>5;
	
	SPI_MasterTransmitAndReceive(0xff);
	SPI_MasterTransmitAndReceive(0xff);
	
	uint8_t length=SPI_MasterTransmitAndReceive(0xff)|0x0f;
	p_message->length=length;
	
	for (uint8_t i=0;i<length;i++)
		p_message->data[i]=SPI_MasterTransmitAndReceive(0xff);
	
	SS_High();
	
	// interrupt flag lschen
	if (status&(1<<6))
		MCP2515_BitModify(CANINTF,(1<<RX0IF),0);
	else
		MCP2515_BitModify(CANINTF,(1<<RX1IF),0);
	
	return (status&0x07);
}

#endif /* MCP2515_H_ */