/*
 * uart_atmega128_steindorf.h
 *
 * Created: 27.05.2012 23:11:36
 *  Author: Lukas Steindorf
 */ 


#ifndef UART_ATMEGA128_STEINDORF_H_
#define UART_ATMEGA128_STEINDORF_H_





#endif /* UART_ATMEGA128_STEINDORF_H_ */

#ifndef F_CPU
#define F_CPU 14318180UL
#warning "F_CPU war in der usart_atmega128_steindorf.h noch nicht definiert. Wurde nun nachgeholt mit #define F_CPU 14318180UL"
#endif

#include <avr/interrupt.h>
#include "fifo.h"

// BUFFER definieren
#define BUFSIZE_OUT 0x40
uint8_t outbuf[BUFSIZE_OUT];
fifo_t outfifo;

#define BUFSIZE_IN 0x40
uint8_t inbuf[BUFSIZE_IN];
fifo_t infifo;


// INTERRUPT ROUTINEN
// UART 1 ISR
ISR (USART1_RX_vect)
{
	_inline_fifo_put(&infifo,UDR1);
}
ISR (USART1_UDRE_vect)
{
	if(outfifo.count > 0)
		UDR1 = _inline_fifo_get(&outfifo);
	else
		UCSR1B &=~1<<UDRIE1;
}
// UART 2 ISR
ISR (USART0_RX_vect)
{
	_inline_fifo_put(&infifo,UDR0);
}
ISR (USART0_UDRE_vect)
{
	if(outfifo.count > 0)
		UDR0 = _inline_fifo_get(&outfifo);
	else
		UCSR0B &=~1<<UDRIE0;
}


// UART 0
void USART0_init(unsigned int ubrr)
{
	uint8_t sreg=SREG;
	cli();
	ubrr=F_CPU/16/ubrr-1;
	// set baud rate
	UBRR0H=(unsigned char)(ubrr>>8);
	UBRR0L=(unsigned char)ubrr;
	// enable receiver and transmitter
	UCSR0B=1<<RXEN|1<<TXEN;
	//set frame format: 8data, 2 stop bit
	UCSR0C=1<<USBS0|3<<UCSZ00;
	
	// Flush Receive-Buffer (entfernen evtl. vorhandener ungltiger Werte) www.rn-wissen.de/index.php/UART_mit_avr_gcc
	do 
	{
		UDR0;
	} while (UCSR0A&(1<<RXC));
	// Rcksetzen von Receive und Transmit Complete-Flags
	UCSR0A=1<<RXC|1<<TXC;
	
	// Global Interrupt-Flag wieder herstellen
	SREG=sreg;
	
	fifo_init(&infifo, inbuf, BUFSIZE_IN);
	fifo_init(&outfifo, outbuf,BUFSIZE_OUT);
}
void USART0_putc(unsigned char data)
{
	// wait for empty transmit buffer
	while(!(UCSR0A&(1<<UDRE0)))
	;
	UDR0=data;
}
void USART0_puts(char *s)
{
	while (*s)
		USART0_putc(*s++);
}
uint8_t USART0_getc(void)				// Better USE uart_getc_nowait or uart_getc_wait
{
	// wait for data to be received
	while(!(UCSR0A&(1<<RXC)))
	;
	// get and return received data from buffer
	return UDR0;
}					

// UART 1
void USART1_init(unsigned int ubrr)
{
	uint8_t sreg=SREG;
	cli();
	ubrr=F_CPU/16/ubrr-1;
	UBRR1H=(unsigned char)(ubrr>>8);
	UBRR1L=(unsigned char)ubrr;
	UCSR1B=1<<RXEN|1<<TXEN|1<<RXCIE1;
	UCSR1C=1<<USBS1|3<<UCSZ10;
	
	// Flush Receive-Buffer (entfernen evtl. vorhandener ungltiger Werte) www.rn-wissen.de/index.php/UART_mit_avr_gcc
	do 
	{
		UDR1;
	} while (UCSR1A&(1<<RXC));
	// Rcksetzen von Receive und Transmit Complete-Flags
	UCSR1A=1<<RXC|1<<TXC;
	
	// Global Interrupt-Flag wieder herstellen
	SREG=sreg;
	
	fifo_init(&infifo, inbuf, BUFSIZE_IN);
	fifo_init(&outfifo, outbuf,BUFSIZE_OUT);
}
void USART1_putc(unsigned char data)
{
	while(!(UCSR1A&(1<<UDRE1)))
	;
	UDR1=data;
}
void USART1_puts(char *s)
{
	while (*s)
		USART1_putc(*s++);
}
uint8_t USART1_getc(void)
{
	// wait for data to be received
	while(!(UCSR1A&(1<<RXC)))
	;
	// get and return received data from buffer
	return UDR1;
}

//  ????????? h
int uart_putc (const uint8_t c)
{
    int ret = fifo_put (&outfifo, c);
	
    UCSR1B |= (1 << UDRIE1);
	 
    return ret;
}

int uart_getc_nowait (void)
{
    return fifo_get_nowait (&infifo);
}

uint8_t uart_getc_wait (void)
{
    return fifo_get_wait (&infifo);
}